"""Learning algorithms on manifolds."""

__all__ = ['TemplateEstimator', 'TemplateClassifier', 'TemplateTransformer']

from ._template import (
    TemplateClassifier, TemplateEstimator, TemplateTransformer
)
