__version__ = '1.19'

import geomstats._backend
import geomstats._logging
import geomstats.geometry.discretized_curves
import geomstats.geometry.euclidean
import geomstats.geometry.hyperbolic
import geomstats.geometry.hypersphere
import geomstats.geometry.invariant_metric
import geomstats.geometry.landmarks
import geomstats.geometry.lie_algebra
import geomstats.geometry.lie_group
import geomstats.geometry.manifold
import geomstats.geometry.minkowski
import geomstats.geometry.poincare_polydisk
import geomstats.geometry.product_manifold
import geomstats.geometry.product_riemannian_metric
import geomstats.geometry.riemannian_metric
import geomstats.geometry.skew_symmetric_matrices
import geomstats.geometry.spd_matrices
import geomstats.geometry.special_euclidean
import geomstats.geometry.special_orthogonal  # NOQA
